from typing import cast

from crewai.agent import Agent
from crewai.agents.agent_builder.base_agent import BaseAgent, PlatformAppOrAction
from crewai.task import Task
from crewai.tools import BaseTool
from crewai.utilities.agent_utils import (
    get_tool_names,
    parse_tools,
    render_text_description_and_args,
)
from crewai.utilities.prompts import Prompts
from crewai.utilities.token_counter_callback import TokenCalcHandler
from pydantic import Field, PrivateAttr, model_validator

from .dbos_agent_executor import DBOSAgentExecutor
from .dbos_llm import DBOSLLM
from .dbos_utils import StepConfig


class DBOSAgent(BaseAgent):
    """Wrap an agent for use in DBOS durable workflows.

    Automatically wraps LLM calls (both llm and function_calling_llm) as DBOS steps.

    Attributes:
            wrapped_agent: The underlying agent instance.
            llm_step_config: The DBOS step configuration to use for LLM steps.
            function_calling_llm_step_config:  The DBOS step configuration to use for function calling LLM steps.
    """

    agent_name: str = Field(description="The unique name of the DBOS agent.")
    orig_agent: Agent = Field(exclude=True, description="The original agent instance.")
    llm_step_config: StepConfig | None = Field(
        default=None,
        description="The DBOS step configuration to use for LLM steps.",
    )
    function_calling_llm_step_config: StepConfig | None = Field(
        default=None,
        description="The DBOS step configuration to use for function calling LLM steps.",
    )
    _wrapped_agent: Agent = PrivateAttr()

    @model_validator(mode="before")
    @classmethod
    def pre_init_setup(cls, values):
        if "orig_agent" not in values or not isinstance(values["orig_agent"], Agent):
            raise ValueError("orig_agent must be provided and be an instance of Agent")
        # populate required fields: role, goal, backstory
        agent = cast(Agent, values["orig_agent"])
        values["role"] = agent.role
        values["goal"] = agent.goal
        values["backstory"] = agent.backstory
        return values

    @model_validator(mode="after")
    def post_init_setup(self):
        # Create a deep copy of the agent to avoid mutating the original
        self._wrapped_agent = self.orig_agent.copy()
        # Replace the original create_agent_executor with a DBOS wrap
        object.__setattr__(
            self._wrapped_agent, "create_agent_executor", self.create_agent_executor
        )
        # Wrap LLM/function calling LLM as steps
        self._wrapped_agent.llm = DBOSLLM(
            orig_llm=self._wrapped_agent.llm,
            step_config=self.llm_step_config,
            name_prefix=f"{self.agent_name}.llm",
        )
        if self._wrapped_agent.function_calling_llm:
            self._wrapped_agent.function_calling_llm = DBOSLLM(
                orig_llm=self._wrapped_agent.function_calling_llm,
                step_config=self.function_calling_llm_step_config,
                name_prefix=f"{self.agent_name}.function_calling_llm",
            )
        return self

    def execute_task(
        self,
        task: Task,
        context: str | None = None,
        tools: list[BaseTool] | None = None,
    ) -> str:
        return self._wrapped_agent.execute_task(
            task=task,
            context=context,
            tools=tools,
        )

    def get_delegation_tools(self, agents: list[BaseAgent]):
        return self._wrapped_agent.get_delegation_tools(agents=agents)

    def get_platform_tools(self, apps: list[PlatformAppOrAction]) -> list[BaseTool]:
        return self._wrapped_agent.get_platform_tools(apps=apps)

    def get_mcp_tools(self, mcps: list[str]) -> list[BaseTool]:
        """Convert MCP server references to CrewAI tools."""
        return self._wrapped_agent.get_mcp_tools(mcps=mcps)

    def create_agent_executor(
        self, tools: list[BaseTool] | None = None, task=None
    ) -> None:
        """Create a DBOS agent executor for the agent.

        Returns:
            An instance of the DBOSAgentExecutor class.
        """
        raw_tools: list[BaseTool] = tools or self._wrapped_agent.tools or []
        parsed_tools = parse_tools(raw_tools)

        prompt = Prompts(
            agent=self._wrapped_agent,
            has_tools=len(raw_tools) > 0,
            i18n=self._wrapped_agent.i18n,
            use_system_prompt=self._wrapped_agent.use_system_prompt,
            system_template=self._wrapped_agent.system_template,
            prompt_template=self._wrapped_agent.prompt_template,
            response_template=self._wrapped_agent.response_template,
        ).task_execution()

        stop_words = [self._wrapped_agent.i18n.slice("observation")]

        if self._wrapped_agent.response_template:
            stop_words.append(
                self._wrapped_agent.response_template.split("{{ .Response }}")[
                    1
                ].strip()
            )

        self._wrapped_agent.agent_executor = DBOSAgentExecutor(
            llm=self._wrapped_agent.llm,
            task=task,
            agent=self._wrapped_agent,
            crew=self._wrapped_agent.crew,
            tools=parsed_tools,
            prompt=prompt,
            original_tools=raw_tools,
            stop_words=stop_words,
            max_iter=self._wrapped_agent.max_iter,
            tools_handler=self._wrapped_agent.tools_handler,
            tools_names=get_tool_names(parsed_tools),
            tools_description=render_text_description_and_args(parsed_tools),
            step_callback=self._wrapped_agent.step_callback,
            function_calling_llm=self._wrapped_agent.function_calling_llm,
            respect_context_window=self._wrapped_agent.respect_context_window,
            request_within_rpm_limit=(
                self._wrapped_agent._rpm_controller.check_or_wait
                if self._wrapped_agent._rpm_controller
                else None
            ),
            callbacks=[TokenCalcHandler(self._wrapped_agent._token_process)],
            name_prefix=self.agent_name,
        )



"""DBOS LLM class for CrewAI.

This module provides the DBOS wrapper class for all LLM implementations
in CrewAI, making call() a DBOS step.
"""

from typing import Any

from crewai.llms.base_llm import BaseLLM
from dbos import DBOS
from pydantic import BaseModel

from .dbos_utils import StepConfig


class DBOSLLM(BaseLLM):
    def __init__(
        self,
        name_prefix: str,
        orig_llm: BaseLLM,
        step_config: StepConfig | None = None,
    ):
        """Initialize the DBOSLLM with an underlying LLM and step configuration.

        Args:
            orig_llm: The original LLM instance to wrap.
            step_config: Optional DBOS step configuration for LLM calls.
        """
        super().__init__(
            model=orig_llm.model,
            temperature=orig_llm.temperature,
            stop=orig_llm.stop,
        )
        self._orig_llm = orig_llm
        self._step_config = step_config or {}

        # Wrap the call method as a DBOS step
        @DBOS.step(name=f"{name_prefix}.{orig_llm.model}", **self._step_config)
        def dbos_call(
            messages: str | list[dict[str, str]],
            tools: list[dict] | None = None,
            callbacks: list[Any] | None = None,
            available_functions: dict[str, Any] | None = None,
            from_task: Any | None = None,
            from_agent: Any | None = None,
            response_model: type[BaseModel] | None = None,
        ) -> str | Any:
            return self._orig_llm.call(
                tools=tools,
                messages=messages,
                callbacks=callbacks,
                available_functions=available_functions,
                from_task=from_task,
                from_agent=from_agent,
                response_model=response_model,
            )

        self._dbos_call = dbos_call

    # Return whatever attribute the underlying LLM has, except for the call method
    def __getattr__(self, name: str) -> Any:
        if name == "call":
            return self._dbos_call
        return getattr(self._orig_llm, name)

    def call(
        self,
        messages: str | list[dict[str, str]],
        tools: list[dict] | None = None,
        callbacks: list[Any] | None = None,
        available_functions: dict[str, Any] | None = None,
        from_task: Any | None = None,
        from_agent: Any | None = None,
        response_model: type[BaseModel] | None = None,
    ) -> str | Any:
        return self._dbos_call(
            messages=messages,
            tools=tools,
            callbacks=callbacks,
            available_functions=available_functions,
            from_task=from_task,
            from_agent=from_agent,
            response_model=response_model,
        )



"""Agent executor for DBOS durable agents.

Handles agent execution flow including LLM interactions, tool execution,
and memory management. Also automatically wraps LLM calls as DBOS steps.
"""

from collections.abc import Callable
from typing import Any, cast

from crewai.agents.agent_builder.base_agent import BaseAgent
from crewai.agents.crew_agent_executor import CrewAgentExecutor
from crewai.agents.tools_handler import ToolsHandler
from crewai.tools.structured_tool import CrewStructuredTool
from dbos import DBOS

from .dbos_llm import DBOSLLM


# Overload invoke with DBOS workflow
@DBOS.workflow(name="DBOSAgentExecutor.invoke")
def dbos_invoke(cai_exec_name: str, inputs: dict[str, str]) -> dict[str, Any]:
    """Invoke the agent executor within a DBOS workflow.
    Args:
        cai_exec_name: The name prefix of the DBOSAgentExecutor instance.
        inputs: Input dictionary containing the task input.
    Returns:
        Dictionary with agent output.
    """
    # TODO: This means DBOS cannot recover workflows if the executor instance doesn't exist.
    cai_exec = _dbos_local_executor_cache.get(cai_exec_name)
    if cai_exec is None:
        raise ValueError(f"DBOSAgentExecutor with name {cai_exec_name} not found.")
    # Get the original invoke method
    orig_exec = cast(CrewAgentExecutor, super(DBOSAgentExecutor, cai_exec))
    return orig_exec.invoke(inputs)


_dbos_local_executor_cache: dict[str, CrewAgentExecutor] = {}


class DBOSAgentExecutor(CrewAgentExecutor):
    """Executor for DBOS agents.

    Manages the execution lifecycle of an agent including prompt formatting,
    LLM interactions, tool execution, and feedback handling. Also automatically wraps LLM calls as DBOS steps.
    """

    def __init__(
        self,
        llm: Any,
        task: Any,
        crew: Any,
        agent: BaseAgent,
        prompt: dict[str, str],
        max_iter: int,
        tools: list[CrewStructuredTool],
        tools_names: str,
        stop_words: list[str],
        tools_description: str,
        tools_handler: ToolsHandler,
        step_callback: Any = None,
        original_tools: list[Any] | None = None,
        function_calling_llm: Any = None,
        respect_context_window: bool = False,
        request_within_rpm_limit: Callable[[], bool] | None = None,
        callbacks: list[Any] | None = None,
        *,
        name_prefix: str,
    ) -> None:
        """Initialize executor.

        Args:
            llm: Language model instance.
            task: Task to execute.
            crew: Crew instance.
            agent: Agent to execute.
            prompt: Prompt templates.
            max_iter: Maximum iterations.
            tools: Available tools.
            tools_names: Tool names string.
            stop_words: Stop word list.
            tools_description: Tool descriptions.
            tools_handler: Tool handler instance.
            step_callback: Optional step callback.
            original_tools: Original tool list.
            function_calling_llm: Optional function calling LLM.
            respect_context_window: Respect context limits.
            request_within_rpm_limit: RPM limit check function.
            callbacks: Optional callbacks list.
            name_prefix: Prefix for DBOS workflow names.
        """
        if not isinstance(llm, DBOSLLM):
            raise ValueError("LLM must be a DBOSLLM instance.")

        if function_calling_llm is not None and not isinstance(
            function_calling_llm, DBOSLLM
        ):
            raise ValueError("function_calling_llm must be a DBOSLLM instance or None.")

        super().__init__(
            llm=llm,
            task=task,
            crew=crew,
            agent=agent,
            prompt=prompt,
            max_iter=max_iter,
            tools=tools,
            tools_names=tools_names,
            stop_words=stop_words,
            tools_description=tools_description,
            tools_handler=tools_handler,
            step_callback=step_callback,
            original_tools=original_tools,
            function_calling_llm=function_calling_llm,
            respect_context_window=respect_context_window,
            request_within_rpm_limit=request_within_rpm_limit,
            callbacks=callbacks,
        )

        self._name_prefix = name_prefix
        _dbos_local_executor_cache[name_prefix] = self

    def invoke(self, inputs: dict[str, str]) -> dict[str, Any]:
        """Invoke the agent executor within a DBOS workflow.

        Args:
            inputs: Input dictionary containing the task input.

        Returns:
            Dictionary with agent output.
        """
        return dbos_invoke(self._name_prefix, inputs)



from typing import TypedDict


class StepConfig(TypedDict, total=False):
    """Configuration for a step in the DBOS workflow."""

    retries_allowed: bool
    interval_seconds: float
    max_attempts: int
    backoff_rate: float
